"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.battery as bat

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class Battery(ElectricalEntity, bat.Battery):
    """
    Extension of pyCity_base class Battery for scheduling purposes.

    Initialize Battery.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    e_el_max : float
        Electric capacity of the battery [kWh].
    p_el_max_charge : float
        Maximum charging power [kW].
    p_el_max_discharge : float, optional
        Maximum discharging power [kW]. Defaults to zero.
    soc_init : float, optional
        Initial state of charge. Defaults to 50%.
    eta : float, optional
        Charging and discharging efficiency. Must be in (0,1].
        Defaults to one.
    storage_end_equality : bool, optional
        `True` if the soc at the end of the scheduling has to be equal to
        the initial soc.
        `False` if it has to be greater or equal than the initial soc.
        Defaults to `False`.

    Notes
    -----
    - Batteries offer sets of constraints for operation. In the `convex` mode
      the following constraints and bounds are generated by the battery:

    .. math::

        p_{el} &=& p_{el\\_demand} - p_{el\\_supply} \\\\
        p_{el\\_max\\_discharge} &\\geq& p_{el\\_supply} \\geq 0 \\\\
        p_{el\\_max\\_charge} &\\geq& p_{el\\_demand} \\geq 0 \\\\
        e_{el\\_max} &\\geq& e_{el} \\geq 0 \\\\
        e_{el} &=& e_{el\\_previous} + (\\eta * p_{el\\_demand}
        - (1 / \\eta) * p_{el\\_supply}) * \\Delta t \\\\
        \\text{with} \\quad e_{el\\_previous} &=& \\
        \\begin{bmatrix} e_{el\\_ini} & e_{el\\_0} & \\cdots & e_{el\\_n-1}\\end{bmatrix}

    - Additional constraints generated by the parameters are:

    .. math::

        e_{el\\_t\\_last} &=& soc\\_init * e_{el\\_max}, & \\quad \\text{if storage_end_equality} \\\\
        e_{el\\_t\\_last} &\\geq& soc\\_init * e_{el\\_max}, & \\quad \\text{else}

    - In `integer` mode the following constraints are added in addition:

    .. math::

        p_{el\\_demand} &\\leq& p_{state} * p_{el\\_max\\_discharge} \\\\
        p_{el\\_supply} &\\leq& (1-p_{state}) * p_{el\\_max\\_charge}
    """

    def __init__(self, environment, e_el_max, p_el_max_charge, p_el_max_discharge=None, soc_init=0.5, eta=0.95,
                 storage_end_equality=False):

        capacity = e_el_max * 3600 * 1000
        soc_abs = soc_init * capacity  # absolute SOC
        super().__init__(environment, soc_abs, capacity, 0.0, eta, eta)
        self._long_id = "BAT_" + self._id_string

        self.objective = 'peak-shaving'
        self.e_el_max = e_el_max
        self.soc_init = soc_init  # relative SOC
        self.p_el_max_charge = p_el_max_charge
        self.p_el_max_discharge = p_el_max_discharge if p_el_max_discharge is not None else p_el_max_charge
        self.eta = eta
        self.storage_end_equality = storage_end_equality

        self.new_var("p_el_demand")
        self.new_var("p_el_supply")
        self.new_var("p_state", dtype=np.bool_, func=Battery._get_p_state_func)
        self.new_var("e_el")

    def _get_p_state_func(self, model):
       return self.schedule["p_el_demand"][self.op_slice] > self.schedule["p_el_supply"][self.op_slice]
        
    def populate_model(self, model, mode="convex"):
        """
        Add device block of variables and constraints to pyomo ConcreteModel.

        Call parent's `populate_model` method and set variables lower bounds to
        `None`. Then add variables for demand, supply and the state of charge,
        with their corresponding upper bounds (`self.p_el_max_charge`,
        `self.p_el_max_discharge`, `self.e_el_max`). Finally add continuity
        constraints to the block.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        super().populate_model(model, mode)
        m = self.model
        if mode in ["convex", "integer"]:
            # additional variables for battery
            m.p_el_demand_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0.0, self.p_el_max_charge),
                                           initialize=0)
            m.p_el_supply_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0.0, self.p_el_max_discharge),
                                           initialize=0)
            m.e_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0.0, self.e_el_max),
                                    initialize=self.e_el_max*self.soc_init)

            def p_el_demand_vars_lb_rule(model, t):
                return model.p_el_demand_vars[t] >= 0.0
            m.p_el_demand_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_demand_vars_lb_rule)

            def p_el_demand_vars_ub_rule(model, t):
                return model.p_el_demand_vars[t] <= self.p_el_max_charge
            if mode != "integer":
                m.p_el_demand_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_demand_vars_ub_rule)

            def p_el_supply_vars_lb_rule(model, t):
                return model.p_el_supply_vars[t] >= 0.0
            m.p_el_supply_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_supply_vars_lb_rule)

            def p_el_supply_vars_ub_rule(model, t):
                return model.p_el_supply_vars[t] <= self.p_el_max_discharge
            if mode != "integer":
                m.p_el_supply_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_supply_vars_ub_rule)

            def e_el_vars_lb_rule(model, t):
                return model.e_el_vars[t] >= 0.0
            m.e_el_vars_lb_constr = pyomo.Constraint(m.t, rule=e_el_vars_lb_rule)

            def e_el_vars_ub_rule(model, t):
                return model.e_el_vars[t] <= self.e_el_max
            m.e_el_vars_ub_constr = pyomo.Constraint(m.t, rule=e_el_vars_ub_rule)

            def p_rule(model, t):
                return model.p_el_vars[t] == model.p_el_demand_vars[t] - model.p_el_supply_vars[t]
            m.p_constr = pyomo.Constraint(m.t, rule=p_rule)
            m.e_el_init = pyomo.Param(default=self.soc_init * self.e_el_max, mutable=True)

            def e_rule(model, t):
                delta = (
                        (self.eta_charge * model.p_el_demand_vars[t]
                         - (1.0 / self.eta_discharge) * model.p_el_supply_vars[t])
                        * self.time_slot
                )
                e_el_last = model.e_el_vars[t-1] if t >= 1 else model.e_el_init
                return model.e_el_vars[t] == e_el_last + delta
            m.e_constr = pyomo.Constraint(m.t, rule=e_rule)

            def e_end_lb_rule(model):
                return model.e_el_vars[self.op_horizon-1] >= self.e_el_max * self.soc_init
            m.e_end_lb_constr = pyomo.Constraint(rule=e_end_lb_rule)

            def e_end_ub_rule(model):
                return model.e_el_vars[self.op_horizon-1] <= self.e_el_max * self.soc_init
            if self.storage_end_equality:
                m.e_end_ub_constr = pyomo.Constraint(rule=e_end_ub_rule)

            if mode == "integer":
                m.p_state_vars = pyomo.Var(m.t, domain=pyomo.Binary, initialize=0)

                def p_el_demand_vars_ub_rule(model, t):
                    return model.p_el_demand_vars[t] <= model.p_state_vars[t] * self.p_el_max_charge
                m.p_el_demand_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_demand_vars_ub_rule)

                def p_el_supply_vars_ub_rule(model, t):
                    return model.p_el_supply_vars[t] <= (1 - model.p_state_vars[t]) * self.p_el_max_discharge
                m.p_el_supply_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_supply_vars_ub_rule)

        else:
            raise ValueError(
                "Mode %s is not implemented by class Battery." % str(mode)
            )
        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        if timestep == 0:
            m.e_el_init = self.soc_init * self.e_el_max
        else:
            m.e_el_init = self.e_el_schedule[timestep-1]
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.demand.electrical_demand as ed

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class CurtailableLoad(ElectricalEntity, ed.ElectricalDemand):
    """
    Extension of pyCity_base class ElectricalDemand for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    p_el_nom : float
        Nominal electric power in [kW].
    max_curtailment : float
        Maximal Curtailment of the load
    max_low : int, optional
        Maximum number of timesteps the curtailable load can stay under
        nominal load
    min_full : int, optional
        Minimum number of timesteps the curtailable load has to stay at
        nominal operation level when switching to the nominal operation
        level

    Notes
    -----
    - CLs offer sets of constraints for operation. In the `convex` mode the following
      constraints and bounds are generated by the CL:

    .. math::
        p_{el\\_nom} * max\\_curtailment \\geq p_{el} \\geq 0 \\\\
        \\sum_{j=i}^{i+max\\_low+min\\_full} p_{el\\_j} \\geq p_{el\\_nom} *
        (min\\_full + max\\_low * max\\_curtailment)

    - The last constraint is replaced in integer mode with the following constraints:

    .. math::
        p_{el} \\geq p_{state} * p_{el\\_nom} \\\\
        \\sum_{j=i}^{i+max\\_low} p_{state\\_j} \\geq 1 \\\\
        \\sum_{j=i}^{i+min\\_full-1} p_{state\\_j} \\geq
        (p_{state\\_i} - p_{state\\_i-1}) * min\\_full

    - These constraints take also the previous values before the current optimization
      horizon into account using the current schedule. Values before :math:`t=0` are
      assumed to be perfect.
    """

    def __init__(self, environment, p_el_nom, max_curtailment, max_low=None, min_full=None):
        shape = environment.timer.timesteps_total
        super().__init__(environment, 0, np.zeros(shape))
        self._long_id = "CUL_" + self._id_string

        self.p_el_nom = p_el_nom
        self.max_curt = max_curtailment
        if max_low is not None or min_full is not None:
            assert max_low is not None
            assert min_full is not None
            assert min_full >= 1
            assert max_low >= 0
        self.max_low = max_low
        self.min_full = min_full
        self.p_el_curt = self.p_el_nom * self.max_curt
        self.new_var("p_state", dtype=np.bool_, func=CurtailableLoad._get_p_state_func)

    def _get_p_state_func(self, model):
        return self.schedule["p_el"][self.op_slice] > 0.99*self.p_el_nom


    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel

        Call parent's `populate_model` method and set variables upper bounds to
        the loadcurve and lower bounds to `self.p_el_Min`.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Uses integer variables for max_low and min_full constraints if necessary
        """
        super(CurtailableLoad, self).populate_model(model, mode)
        m = self.model

        if mode in ["convex", "integer"]:

            def p_el_vars_lb_curt_rule(model, t):
                return model.p_el_vars[t] >= self.p_el_curt

            def p_el_vars_lb_nom_rule(model, t):
                return model.p_el_vars[t] >= self.p_el_nom

            def p_el_vars_ub_rule(model, t):
                return model.p_el_vars[t] <= self.p_el_nom

            m.p_el_vars_lb_curt_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_curt_rule)
            m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)
            if self.max_low is None:
                # if max_low is not set the entity can choose p_state freely.
                # as a result no constraints are required
                pass
            elif self.max_low == 0:
                # if max_low is zero the p_state_vars would have to always be one
                # this results in operation at always 100%.
                # the following bound is enough to represent this behaviour
                m.p_el_vars_lb_nom_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_nom_rule)
            elif mode == "integer":
                # generate integer constraints for max_low min_full values

                # create binary variables representing the state if the device is operating at full level
                m.p_state_vars = pyomo.Var(m.t, domain=pyomo.Binary, initialize=0)

                # coupling the state variable to the electrical variable
                # since operation infinitely close to 100% can be chosen by the entity to create a state
                # of zero, coupling in one direction is sufficient.

                def p_state_rule(model, t):
                    return model.p_state_vars[t] * self.p_el_nom <= model.p_el_vars[t]
                m.p_state_integer_constr = pyomo.Constraint(m.t, rule=p_state_rule)

                # creat constraints which can be used by update_model to take previous states into account.
                # update_schedule only needs to modify RHS which should be faster than deleting and creating
                # new constraints

                max_overlap = max(self.max_low, self.min_full)

                # add constraints forcing the entity to operate at least once at 100% between every range
                # of max_low + 1 in the op_horizon
                m.previous_n_states = pyomo.Param(pyomo.RangeSet(1, max_overlap), mutable=True,
                                                  initialize=True)

                def p_full_rule(model, t):
                    e = 0
                    entries = 0
                    for t_backwards in range(t-self.max_low, t + 1, 1):
                        if t_backwards >= 0:
                            e += model.p_state_vars[t_backwards]
                        else:
                            e += model.previous_n_states[-t_backwards]
                        entries += 1

                    assert entries == self.max_low + 1
                    return e >= 1
                m.P_full_integer_constr = pyomo.Constraint(m.t, rule=p_full_rule)

                # add constraints to operate at a minimum of min_full timesteps at 100% when switching
                # from the state 0 to the state 1
                if self.min_full > 1:
                    def p_on_rule(model, t):
                        e = 0
                        entries = 0
                        for t_backwards in range(t - self.min_full + 1, t + 1, 1):
                            if t_backwards >= self.op_horizon:
                                e += 1
                            elif t_backwards >= 0:
                                e += model.p_state_vars[t_backwards]
                            else:
                                e += model.previous_n_states[-t_backwards]
                            entries += 1
                        assert entries == self.min_full
                        t_flank_start = t - self.min_full
                        t_flank_end = t - self.min_full + 1
                        if t_flank_start >= 0:
                            flank_start = model.p_state_vars[t_flank_start]
                        else:
                            flank_start = model.previous_n_states[-t_flank_start]
                        if t_flank_end >= 0:
                            flank_end = model.p_state_vars[t_flank_end]
                        else:
                            flank_end = model.previous_n_states[-t_flank_end]
                        return (flank_end - flank_start) * self.min_full <= e

                    m.P_on_integer_constr = pyomo.Constraint(pyomo.RangeSet(0,
                                                                            self.op_horizon + self.min_full - 2),
                                                             rule=p_on_rule)

            else:
                # generate relaxed constraints with max_low min_full values
                width = self.min_full + self.max_low

                m.previous_n_consumptions = pyomo.Param(pyomo.RangeSet(1, width-1),
                                                        mutable=True,
                                                        initialize=self.p_el_nom)

                def p_average_rule(model, t):
                    e = 0
                    entries = 0
                    for t_backwards in range(t - width + 1, t + 1, 1):
                        if t_backwards >= 0:
                            e += model.p_el_vars[t_backwards]
                        else:
                            e += model.previous_n_consumptions[-t_backwards]
                        entries += 1
                    assert entries == width
                    return e >= self.p_el_nom * self.min_full + self.p_el_curt * self.max_low

                m.P_average_constr = pyomo.Constraint(m.t, rule=p_average_rule)

        else:
            raise ValueError(
                "Mode %s is not implemented by class CurtailableLoad." % str(mode)
            )
        return

    def update_model(self, mode="convex"):
        super(CurtailableLoad, self).update_model(mode)
        m = self.model
        timestep = self.timer.current_timestep
        if hasattr(m, "previous_n_states"):

            for key in m.previous_n_states.sparse_iterkeys():
                if timestep - key >= 0:
                    previous_value = self.p_state_schedule[timestep - key] * 1.0
                else:
                    # for timesteps below zero a perfect state is assumed.
                    previous_value = True
                m.previous_n_states[key] = previous_value
        if hasattr(m, "previous_n_consumptions"):
            for key in m.previous_n_consumptions.sparse_iterkeys():
                if timestep - key >= 0:
                    previous_value = self.p_el_schedule[timestep - key]
                else:
                    # for timesteps below zero a perfect state is assumed.
                    previous_value = self.p_el_nom

                m.previous_n_consumptions[key] = previous_value
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

import pycity_scheduling.classes as classes
from pycity_scheduling.algorithms import CentralOptimization
from pycity_scheduling.util import factory
from pycity_scheduling.util.metric import calculate_costs


# This is a simple power scheduling example combined with a Robust Optimization (RO) application.


def main(do_plot=False):
    print("\n\n------ Example 24: Scheduling Robust Optimization ------\n\n")

    # Use a simple environment of 6 hours with quarter-hourly resolution (=15min=900sec):
    env = factory.generate_standard_environment(step_size=900, op_horizon=6)

    # Make it "attractive" for the customer to shift demand into the first half of the scheduling period
    # (compare example_23_post-processing_metrics_evaluation.py):
    env.prices.tou_prices = np.array([5]*3 + [10]*3)

    district = classes.CityDistrict(env)
    # The sample building in this example comes with a constant space heating load of 10kW, thermal storage of capacity
    # 20kWh and an electric heater of thermal nominal power 20kW:
    bd = factory.generate_simple_building(env, sh=10, ths=20, eh=20)

    district.addEntity(bd, (0, 0))

    # Perform the scheduling without RO:
    opt = CentralOptimization(district, robustness=(6, 0.0))
    opt.solve(robustness=(6, 0.0))
    bd.copy_schedule("ref")

    # Protect 6 time steps and assume a maximum deviation of 50% in each time step.
    # Such a high deviation is usually unrealistic, but makes it a good example here.
    opt.solve(robustness=(6, 0.5))

    # Print schedules/results:
    # We compare the schedules from the two performed schedulings (with/without RO) with each other.
    np.set_printoptions(formatter={'float': '{: >8.3f}'.format})
    print('Building p_el:')
    print(bd.p_el_ref_schedule)
    print(bd.p_el_schedule)
    print('ThermalEnergyStorage e_th_heat:')
    print(bd.bes.tes_units[0].e_th_heat_ref_schedule)
    print(bd.bes.tes_units[0].e_th_heat_schedule)
    print('ThermalEnergyStorage Limits:')
    print(list(bd.model.lower_robustness_bounds[:].value))
    print(list(bd.model.upper_robustness_bounds[:].value))
    print('ElectricHeater p_th_heat:')
    print(bd.bes.electrical_heaters[0].p_th_heat_ref_schedule)
    print(bd.bes.electrical_heaters[0].p_th_heat_schedule)
    print('SpaceHeating p_th_heat:')
    print(bd.apartments[0].th_heating_demand_list[0].p_th_heat_ref_schedule)
    print(bd.apartments[0].th_heating_demand_list[0].p_th_heat_schedule)
    print('Costs:')
    bd.load_schedule("ref")
    print('{:.2f}'.format(calculate_costs(bd)))
    bd.load_schedule("default")
    print('{:.2f}'.format(calculate_costs(bd)))
    return


# Conclusions:
# If Robust Optimization (RO) is applied, the flexibility of the thermal energy storage is not fully used. This is best
# seen in the first four time steps, where the state-of-charge (SOC) is lower than without the robust approach. Instead,
# the 'energy difference' is used to cater for uncertainties that may stem from an uncertain thermal demand of the
# building. As a trade-off, the RO schedule becomes always less optimal (by means of the objective value) than without
# RO. This becomes evident by the higher costs for the robust case compared to the non-robust case in this example.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a very simple power scheduling example using the central optimization algorithm to demonstrate the impact
# of system level objective "peak-shaving".


def main(do_plot=False):
    print("\n\n------ Example 11: Objective Peak-Shaving ------\n\n")

    # Define timer, price, weather, and environment objects:
    t = Timer(op_horizon=96, step_size=900, initial_date=(2015, 4, 1))
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "peak-shaving":
    cd = CityDistrict(environment=e, objective='peak-shaving')

    # Schedule some sample buildings. The buildings' objectives are defined as "none".
    n = 10
    for i in range(n):
        bd = Building(environment=e, objective='none')
        cd.addEntity(entity=bd, position=[0, i])
        bes = BuildingEnergySystem(environment=e)
        bd.addEntity(bes)
        ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
        bes.addDevice(ths)
        eh = ElectricHeater(environment=e, p_th_nom=10)
        bes.addDevice(eh)
        ap = Apartment(environment=e)
        bd.addEntity(ap)
        fi = FixedLoad(e, method=1, annual_demand=3000.0, profile_type='H0')
        ap.addEntity(fi)
        sh = SpaceHeating(environment=e, method=1, living_area=120, specific_demand=90, profile_type='HEF')
        ap.addEntity(sh)
        pv = Photovoltaic(environment=e, method=1, peak_power=8.2)
        bes.addDevice(pv)
        bat = Battery(environment=e, e_el_max=12.0, p_el_max_charge=4.6, p_el_max_discharge=4.6, eta=1.0)
        bes.addDevice(bat)

    # Perform the scheduling:
    opt = CentralOptimization(city_district=cd)
    opt.solve()
    cd.copy_schedule("peak-shaving")

    # Print and show the city district's schedule:
    print("Schedule of the city district:")
    print(list(cd.p_el_schedule))
    plt.plot(cd.p_el_schedule)
    plt.ylim([-2.0, 5.0])
    plt.xlabel('Time in hours')
    plt.ylabel('Electrical power in kW')
    plt.title('City district scheduling result')
    plt.grid()
    if do_plot:
        plt.show()
    return


# Conclusions:
# Using "peak-shaving" as the system level objective results in a "flat" power profile for the considered city district.
# In other words, this means that power peaks (caused by peak demands) and power valleys (caused by PV feed-in) are
# mitigated using the local flexibility of the buildings' assets. A flat power profile is usually the preferred system
# operation from the viewpoint of the network operator.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes import (CityDistrict, Building, Photovoltaic, WindEnergyConverter)
from pycity_scheduling.algorithms.algorithm import IterationAlgorithm, DistributedAlgorithm, SolverNode
from pycity_scheduling.solvers import DEFAULT_SOLVER, DEFAULT_SOLVER_OPTIONS


class DualDecomposition(IterationAlgorithm, DistributedAlgorithm):
    """
    Implementation of the distributed Dual Decomposition algorithm as described in [1].

    Parameters
    ----------
    city_district : CityDistrict
    solver : str, optional
        Solver to use for solving (sub)problems.
    solver_options : dict, optional
        Options to pass to calls to the solver. Keys are the name of
        the functions being called and are one of `__call__`, `set_instance_`,
        `solve`.
        `__call__` is the function being called when generating an instance
        with the pyomo SolverFactory. In addition to the options provided,
        `node_ids` is passed to this call containing the IDs of the entities
        being optimized.
        `set_instance` is called when a pyomo Model is set as an instance of
        a persistent solver. `solve` is called to perform an optimization. If
        not set, `save_results` and `load_solutions` may be set to false to
        provide a speedup.
    mode : str, optional
        Specifies which set of constraints to use.
        - `convex`  : Use linear constraints
        - `integer`  : May use non-linear constraints
    eps_primal : float, optional
        Primal stopping criterion for the dual decomposition algorithm.
    rho : float, optional
        Step size for the dual decomposition algorithm.
    max_iterations : int, optional
        Maximum number of ADMM iterations.
    robustness : tuple, optional
        Tuple of two floats. First entry defines how many time steps are
        protected from deviations. Second entry defines the magnitude of
        deviations which are considered.

    References
    ----------
    [1] "Distributed Optimization for Scheduling Electrical Demand in Complex City Districts" by
    M. Diekerhof, S. Schwarz, F. Martin, A. Monti in IEEE Systems Journal, vol. 12, no. 4, pp. 3226-3237, Dec. 2018.
    """
    def __init__(self, city_district, solver=DEFAULT_SOLVER, solver_options=DEFAULT_SOLVER_OPTIONS, mode="convex",
                 eps_primal=0.1, rho=2.0, max_iterations=10000, robustness=None):
        super(DualDecomposition, self).__init__(city_district, solver, solver_options, mode)
        self.eps_primal = eps_primal
        self.rho = rho
        self.max_iterations = max_iterations
        self.op_horizon = self.city_district.op_horizon

        # Only consider entities of type CityDistrict, Building, Photovoltaic, WindEnergyConverter
        self._entities = [entity for entity in self.entities if
                          isinstance(entity, (CityDistrict, Building, Photovoltaic, WindEnergyConverter))]

        # Create a solver node for each entity
        self.nodes = [SolverNode(solver, solver_options, [entity], mode, robustness=robustness)
                      for entity in self._entities]

        # Create pyomo parameters for each entity
        for node, entity in zip(self.nodes, self._entities):
            node.model.beta = pyomo.Param(mutable=True, initialize=1)
            node.model.lambdas = pyomo.Param(entity.model.t, mutable=True, initialize=0)
        self._add_objective()

    def _add_objective(self):
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            obj = node.model.beta * entity.get_objective()
            if i == 0:
                # penalty term is expanded and constant is omitted
                # invert sign of p_el_schedule and p_el_vars (omitted for quadratic term)
                for t in range(self.op_horizon):
                    obj -= node.model.lambdas[t] * entity.model.p_el_vars[t]
            else:
                for t in range(self.op_horizon):
                    obj += node.model.lambdas[t] * entity.model.p_el_vars[t]
            node.model.o = pyomo.Objective(expr=obj)
        return

    def _get_objective(self):
        obj_value = 0
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            obj_value += pyomo.value(entity.get_objective())
        return obj_value

    def _presolve(self, full_update, beta, robustness, debug):
        results, params = super()._presolve(full_update, beta, robustness, debug)
        for node, entity in zip(self.nodes, self._entities):
            node.model.beta = self._get_beta(params, entity)
            if full_update:
                node.full_update(robustness)
        results["r_norms"] = []
        results["lambdas"] = []
        return results, params

    def _is_last_iteration(self, results, params, debug):
        if super(DualDecomposition, self)._is_last_iteration(results, params, debug):
            return True
        return results["r_norms"][-1] <= self.eps_primal

    def _iteration(self, results, params, debug):
        super(DualDecomposition, self)._iteration(results, params, debug)
        if "lambdas" not in params:
            params["lambdas"] = np.zeros(self.op_horizon)
        lambdas = params["lambdas"]

        # ------------------------------------------
        # 1) Optimize all entities
        # ------------------------------------------
        for i, node, entity in zip(range(len(self._entities)), self.nodes, self._entities):
            for t in range(self.op_horizon):
                node.model.lambdas[t] = lambdas[t]
            node.obj_update()
            node.solve(debug=debug)

        # ------------------------------------------
        # 2) Calculate incentive signal update
        # ------------------------------------------
        p_el_schedules = np.stack([entity.p_el_schedule for entity in self._entities], axis=0)
        lambdas -= self.rho * p_el_schedules[0]
        lambdas += self.rho * np.sum(p_el_schedules[1:], axis=0)

        # ------------------------------------------
        # 3) Calculate parameters for stopping criteria
        # ------------------------------------------
        r = np.zeros(self.op_horizon)
        r -= p_el_schedules[0]
        r += np.sum(p_el_schedules[1:], axis=0)
        results["r_norms"].append(np.linalg.norm(r, np.inf))
        results["lambdas"].append(np.copy(lambdas))
        results["obj_value"].append(self._get_objective())

        # ------------------------------------------
        # 4) Save required parameters for another iteration
        # ------------------------------------------
        params["lambdas"] = lambdas
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.wind_energy_converter as wec

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class WindEnergyConverter(ElectricalEntity, wec.WindEnergyConverter):
    """
    Extension of pyCity_base class WindEnergyConverter for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    velocity : numpy.ndarray
        Wind speeds in [m/s].
    power : numpy.ndarray
        Power for given velocities in [kW].
    hub_height : float, optional
        Height of the wind energy converter in [m].
    roughness : float, optional
        Roughness of landscape in [m].
    force_renewables : bool, optional
        `True` if generation may not be reduced for optimization purposes.


    Notes
    -----
    - The following constraint is added for removing the bounds from EE:

    .. math::
        p_{el} &=& -p_{el\\_supply}, & \\quad \\text{if force_renewables} \\\\
        0 \\geq p_{el} &\\geq& -p_{el\\_supply} , & \\quad \\text{else}
    """

    def __init__(self, environment, velocity, power,
                 hub_height=70, roughness=0.1, force_renewables=True):
        super(WindEnergyConverter, self).__init__(environment, velocity, power, hub_height, roughness)
        self._long_id = "WEC_" + self._id_string

        self.force_renewables = force_renewables
        self.p_el_supply = self._get_power_supply()

    def _get_power_supply(self):
        # Base class cannot compute values for the whole simulation horizon
        wheather_forecast = self.environment.weather.getWeatherForecast
        (full_wind,) = wheather_forecast(getVWind=True)
        ts = self.timer.time_in_year(from_init=True)
        total_wind = full_wind[ts:ts + self.simu_horizon]
        log_wind = self._logWindProfile(total_wind)
        return np.interp(log_wind, self.velocity, self.power, right=0)

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model
        timestep = self.timestep

        def p_el_vars_lb_rule(model, t):
            return model.p_el_vars[t] >= -self.p_el_supply[timestep + t]

        m.p_el_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_rule)

        def p_el_vars_ub_rule(model, t):
            if self.force_renewables:
                return model.p_el_vars[t] <= -self.p_el_supply[timestep + t]
            else:
                return model.p_el_vars[t] <= 0.0

        m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        p_el_vars_lb_constr_active = m.p_el_vars_lb_constr.active

        def p_el_vars_lb_rule(model, t):
            return model.p_el_vars[t] >= -self.p_el_supply[timestep + t]
        try:
            m.del_component(m.p_el_vars_lb_constr)
        except AttributeError:
            pass
        m.p_el_vars_lb_constr = pyomo.Constraint(m.t, rule=p_el_vars_lb_rule)

        if p_el_vars_lb_constr_active is False:
            m.p_el_vars_lb_constr.deactivate()

        p_el_vars_ub_constr_active = m.p_el_vars_ub_constr.active

        def p_el_vars_ub_rule(model, t):
            if self.force_renewables:
                return model.p_el_vars[t] <= -self.p_el_supply[timestep + t]
            else:
                return model.p_el_vars[t] <= 0.0
        try:
            m.del_component(m.p_el_vars_ub_constr)
        except AttributeError:
            pass
        m.p_el_vars_ub_constr = pyomo.Constraint(m.t, rule=p_el_vars_ub_rule)

        if p_el_vars_ub_constr_active is False:
            m.p_el_vars_ub_constr.deactivate()
        return

    def get_objective(self, coeff=1):
        """
        Objective function of the WindEnergyConverter.

        Return the objective function of the wind energy converter weighted
        with `coeff`. Depending on `self.force_renewables` leave objective
        function empty or build quadratic objective function to minimize
        discrepancy between available power and produced power.

        Parameters
        ----------
        coeff : float, optional
            Coefficient for the objective function.

        Returns
        -------
        ExpressionBase :
            Objective function.
        """
        m = self.model

        s = 0
        for t in self.op_time_vec:
            s += m.p_el_vars[t] * m.p_el_vars[t]
            s -= 2 * self.p_el_supply[self.op_slice][t] * m.p_el_vars[t]
        return coeff * s

    def get_decision_var(self):
        return self.model.p_el_vars



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

import pycity_scheduling.util.factory as factory
from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a simple power scheduling example to demonstrate the integration and interaction of PV and battery storage
# systems using the central optimization algorithm. The scheduling period is two weeks.


def main(do_plot=False):
    print("\n\n------ Example 21: Scheduling PV+Battery System ------\n\n")

    # Scheduling will be performed for one month:
    env = factory.generate_standard_environment(step_size=3600, op_horizon=24*14, mpc_horizon=None,
                                                mpc_step_width=None, initial_date=(2018, 3, 1), initial_time=(0, 0, 0))

    # City district / district operator objective is peak-shaving:
    cd = CityDistrict(environment=env, objective='peak-shaving')

    # Building equipped with an inflexible load and a PV+battery system / building objective is peak-shaving:
    bd = Building(environment=env, objective='peak-shaving')
    cd.addEntity(entity=bd, position=[0, 0])
    ap = Apartment(environment=env)
    bd.addEntity(ap)
    fl = FixedLoad(environment=env, method=1, annual_demand=3500.0, profile_type="H0")
    ap.addEntity(fl)
    bes = BuildingEnergySystem(environment=env)
    bd.addEntity(bes)
    pv = Photovoltaic(environment=env, method=0, area=25.0, beta=30.0, eta_noct=0.15)
    bes.addDevice(pv)
    bat = Battery(environment=env, e_el_max=13.6, p_el_max_charge=24.0, p_el_max_discharge=3.6, soc_init=0.5, eta=1.0,
                  storage_end_equality=True)
    bes.addDevice(bat)

    # Perform the scheduling:
    opt = CentralOptimization(city_district=cd)
    results = opt.solve()
    cd.copy_schedule("central")

    # Plot the (thermal) schedules of interest:
    plot_time = list(range(env.timer.timesteps_used_horizon))

    gs = gridspec.GridSpec(5, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(plot_time, bd.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.ylim([-15, 15])
    plt.grid()
    plt.title("Schedules")
    plt.ylabel("Building [kW]")

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(plot_time, pv.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("PV [kW]")

    ax2 = plt.subplot(gs[2], sharex=ax0)
    ax2.plot(plot_time, fl.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Load [kW]")

    ax3 = plt.subplot(gs[3], sharex=ax0)
    ax3.plot(plot_time, bat.p_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Battery [kW]")

    ax4 = plt.subplot(gs[4], sharex=ax0)
    ax4.plot(plot_time, bat.e_el_schedule)
    plt.xlim((0, env.timer.timesteps_used_horizon - 1))
    plt.grid()
    plt.ylabel("Battery [kWh]")

    plt.xlabel("Time", fontsize=12)

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return


# Conclusions:
# To satisfy the desired peak-shaving objectives of the local building and the city district, it becomes evident that
# the battery storage unit is scheduled in a way that power is charged during times of a PV generation surplus (i.e.,
# noon time). Vice versa, it is discharged during times of a PV generation shortage (i.e., during the night). Moreover,
# the battery storage unit is used to better cope with the fluctuating power consumption/generation inside the building.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
Python class to analyze the generated city district in terms of the total generated and consumed power
"""
import numpy as np
import pyomo.environ as pyomo
import matplotlib.pyplot as plt
from pycity_scheduling.classes import *


class DistrictAnalyzer:
    """
    Python class to analyze the generated city district in terms of the total generated and consumed power
    """
    figure_number = 5

    def __init__(self, city_district):
        self.city_district = city_district
        self.op_horizon = city_district.op_horizon
        self.initial_time_step = self.city_district.environment.timer.time_in_year()
        self.end_time_step = self.initial_time_step + self.op_horizon - 1
        self.PVPower = self._get_total_pv_generation()
        self.FLPower = self._get_total_fixed_load()
        self.HeatDemand, self.FossileHeat, self.ElectricalHeat,\
            self.ElectricalPowerForHeat = self._get_heat_demand()
        self.CHPGeneration = self._get_total_chp_generation()
        self.DLSchedule = np.zeros(self.op_horizon)
        self.BatSchedule = np.zeros(self.op_horizon)
        self.EVSchedule = np.zeros(self.op_horizon)
        self.CHPSchedule = np.zeros(self.op_horizon)
        self.EHSchedule = np.zeros(self.op_horizon)
        self.HPSchedule = np.zeros(self.op_horizon)
        self.final_demand_schedule = np.zeros(self.op_horizon)
        self.final_generation_schedule = np.zeros(self.op_horizon)
        self.district_schedule = np.zeros(self.op_horizon)

    def _get_total_pv_generation(self):
        """
        Function that returns the total PV generated power of the city district in kW per time step
        """
        annual_power = np.zeros(self.city_district.environment.timer.timesteps_horizon)
        total_horizon_power = np.zeros(self.op_horizon)
        for en in self.city_district.get_all_entities():
            if isinstance(en, Photovoltaic):
                annual_power = np.add(annual_power, en.getPower())
        for i in range(self.op_horizon):
            total_horizon_power[i] += annual_power[self.initial_time_step + i] / 1000
        return total_horizon_power

    def _get_total_fixed_load(self):
        """
        Function that returns the total electrical power demand caused by fixed loads in kW per time step
        """
        total_fixed_load = np.zeros(self.op_horizon)
        for en in self.city_district.get_all_entities():
            if isinstance(en, FixedLoad):
                total_fixed_load = np.add(total_fixed_load, en.p_el_schedule)
        return total_fixed_load

    def _get_total_deferrable_load(self):
        """
        Function that returns the total electrical energy demand caused by deferrable loads in kWh
        """
        total_deferrable_load = 0
        for en in self.city_district.get_all_entities():
            if isinstance(en, DeferrableLoad):
                total_deferrable_load += en.e_consumption
        return total_deferrable_load

    def _get_total_battery_capacity(self):
        """
        Function that returns the total storage capacities of classical batteries and electric vehicles in kWh.
        Note, at the end of a simultated period, the soc has to be grater or equal to the initial soc per default.
        Therefore, the calculated total battery capacity in this function is a maximum!
        """
        total_battery_capacity = 0
        for en in self.city_district.get_all_entities():
            if isinstance(en, Battery):
                total_battery_capacity += en.e_el_max*(1-en.soc_init)
        return total_battery_capacity

    def _get_heat_demand(self):
        """
        Function that returns the total heat demand in kWh per time step and the corresponding electrical demand
        in kWh per time step
        """
        total_heat_demand = np.zeros(self.op_horizon)
        total_fossil_generated_heat = np.zeros(self.op_horizon)
        total_electrical_generated_heat = np.zeros(self.op_horizon)
        total_el_power_for_heat = np.zeros(self.op_horizon)
        for building in self.city_district.get_lower_entities():
            building_heat_demand = np.zeros(self.op_horizon)
            eta = 0
            cop = np.zeros(self.op_horizon)
            device = ''
            for en in building.get_all_entities():
                if isinstance(en, SpaceHeating):
                    total_heat_demand = np.add(total_heat_demand, en.p_th_heat_schedule)
                    building_heat_demand = en.p_th_heat_schedule
                if isinstance(en, ElectricHeater):
                    device = 'EH'
                    eta = en.eta
                if isinstance(en, HeatPump):
                    device = 'HP'
                    cop = en.cop
                if isinstance(en, CombinedHeatPower):
                    device = 'CHP'
            if device == 'EH':
                total_electrical_generated_heat = np.add(total_electrical_generated_heat, building_heat_demand)
                building_el_heat_demand = building_heat_demand*eta
            elif device == 'HP':
                total_electrical_generated_heat = np.add(total_electrical_generated_heat, building_heat_demand)
                building_el_heat_demand = np.multiply(building_heat_demand, 1/cop)
            elif device == 'CHP' or device == 'BL':
                total_fossil_generated_heat = np.add(total_fossil_generated_heat, building_heat_demand)
                building_el_heat_demand = np.zeros(self.op_horizon)
            else:
                building_el_heat_demand = np.zeros(self.op_horizon)
                print("Device not listed")
            total_el_power_for_heat = np.add(total_el_power_for_heat, building_el_heat_demand)

        return total_heat_demand, total_fossil_generated_heat, total_electrical_generated_heat, total_el_power_for_heat

    def _get_total_chp_generation(self):
        """
        Pre analyze
        """
        total_chp_generation = np.zeros(self.op_horizon)
        for en in self.city_district.get_all_entities():
            p_el_chp = 0
            power = np.zeros(self.op_horizon)
            if isinstance(en, CombinedHeatPower):
                # Estimation: CHP is gonna run with 80 % of the nominal power in average
                p_el_chp = en.p_el_nom * 0.8
                for i in range(self.op_horizon):
                    power[i] = p_el_chp
            total_chp_generation = np.add(total_chp_generation, power)
        return total_chp_generation

    def plot_electrical_generation_profile(self, plot_now=False):
        """
        Pre analyze
        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.PVPower, label='PV', linestyle='dashed')
        plt.plot(self.CHPGeneration, label='CHP', linestyle='dashed')
        plt.plot(self.PVPower+self.CHPGeneration, label='total')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Generation Profile')
        plt.grid()
        if plot_now == True:
            plt.show()
        return

    def plot_electrical_demand_profile(self, plot_now=False):
        """
        Pre analyze
        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.FLPower+self.ElectricalPowerForHeat, label='Total demand')
        plt.plot(self.FLPower, label='FL', linestyle='dashed')
        plt.plot(self.ElectricalPowerForHeat, label='Electrical Power for Heat', linestyle='dashed')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Demand Profile')
        plt.grid()
        if plot_now == True:
            plt.show()

    def plot_heat_demand_profile(self, plot_now=False):
        """
        Pre analyze
        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.HeatDemand, label='Heat demand')
        plt.plot(self.FossileHeat, label='Fossil generated heat', linestyle='dashed')
        plt.plot(self.ElectricalHeat, label='Electrical generated heat', linestyle='dashed')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Heat Demand and Generation Profile')
        plt.grid()
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.ElectricalPowerForHeat, label='ELectrical power')
        plt.plot(self.ElectricalHeat, label='Electrical generated heat')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Power Demand for Electrically Generated Heat')
        plt.grid()
        if plot_now == True:
            plt.show()
        return

    def plot_electrical_power_imbalance(self, plot_now=False):
        """
        Pre analyze
        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(-self.PVPower-self.CHPGeneration, label='Electrical Power Generation', linestyle='dashed')
        plt.plot(self.FLPower+self.ElectricalPowerForHeat, label='Electrical Power Demand', linestyle='dashed')
        plt.plot(-self.PVPower-self.CHPGeneration+self.FLPower+self.ElectricalPowerForHeat,
                 label='Power Imbalance')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Power Imbalance')
        plt.grid()
        if plot_now == True:
            plt.show()
        return

    def complete_pre_analyze(self):
        """
        Function to plot an estimation of the  city districts schedules if no optimization was done.
        """
        total_generation = self.PVPower + self.CHPGeneration
        total_load = self.FLPower + self.ElectricalPowerForHeat
        total_flexibility = self._get_total_deferrable_load() + self._get_total_battery_capacity()
        imbalance = np.subtract(total_load, total_generation)
        net_imbalance = 0
        for i in range(self.op_horizon):
            net_imbalance += imbalance[i]
        print("net-imbalance: ", net_imbalance)
        print("Flexibility_potential", total_flexibility)

        self.plot_heat_demand_profile()
        self.plot_electrical_demand_profile()
        self.plot_electrical_generation_profile()
        self.plot_electrical_power_imbalance()
        plt.show()
        return

    def add_vectors(self, numpy_vector, pyomo_vector):
        """
        Function to add an indexed pyomo variable with a numpy array elementwise
        """
        for i in range(self.op_horizon):
            numpy_vector[i] += pyomo.value(pyomo_vector[i])
        return numpy_vector

    def convert_pyomo_to_numpy(self, pyomo_vector):
        """
        Function that converts an indexed pyomo variable to a numpy array
        """
        temp = np.zeros(self.op_horizon)
        for i in range(self.op_horizon):
            temp[i] += pyomo.value(pyomo_vector[i])
        return temp

    def _get_flexible_device_schedules(self):
        """
        Function that calculates and saves the aggregated schedules for the flexible entity types DL, Bat, EV
        """
        DL_schedule = np.zeros(self.op_horizon)
        Bat_schedule = np.zeros(self.op_horizon)
        EV_schedule = np.zeros(self.op_horizon)
        for en in self.city_district.get_all_entities():
            if isinstance(en, DeferrableLoad):
                DL_schedule = self.add_vectors(DL_schedule, en.model.p_el_vars)
            if isinstance(en, Battery) and isinstance(en, ElectricVehicle):
                EV_schedule = self.add_vectors(EV_schedule, en.model.p_el_vars)
            if isinstance(en, Battery) and not isinstance(en, ElectricVehicle):
                Bat_schedule = self.add_vectors(Bat_schedule, en.model.p_el_vars)
        self.DLSchedule = DL_schedule
        self.BatSchedule = Bat_schedule
        self.EVSchedule = EV_schedule
        return

    def _get_heat_device_schedules(self):
        """
        Function that calculates and saves the aggregated schedules for the heating entities CHP, EH, HP
        """
        CHP_schedule = np.zeros(self.op_horizon)
        EH_schedule = np.zeros(self.op_horizon)
        HP_schedule = np.zeros(self.op_horizon)
        for en in self.city_district.get_all_entities():
            if isinstance(en, CombinedHeatPower):
                CHP_schedule = self.add_vectors(CHP_schedule, en.model.p_el_vars)
            if isinstance(en, ElectricHeater):
                EH_schedule = self.add_vectors(EH_schedule, en.model.p_el_vars)
            if isinstance(en, HeatPump):
                HP_schedule = self.add_vectors(HP_schedule, en.model.p_el_vars)
        self.CHPSchedule = CHP_schedule
        self.EHSchedule = EH_schedule
        self.HPSchedule = HP_schedule
        return

    def plot_flexible_device_schedules(self, plot_now=False):
        self._get_flexible_device_schedules()
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.DLSchedule, label='DL', linestyle='dashed')
        plt.plot(self.EVSchedule, label='EV', linestyle='dashed')
        plt.plot(self.BatSchedule, label='Bat', linestyle='dashed')
        plt.plot(self.BatSchedule+self.DLSchedule+self.EVSchedule, label='total')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Flexible Devices Profile')
        plt.grid()
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.BatSchedule+self.DLSchedule+self.EVSchedule, label='total')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Aggregated Flexible Devices Profile')
        plt.grid()
        if plot_now:
            plt.show()
        return

    def _extract_demand_and_supply(self, schedule):
        """
        Function to seperate the schedules of Batteries (including Evs) into schedules that contain the demanding and
        the supplying time-steps of flexible loads
        """
        demand_schedule = np.zeros(self.op_horizon)
        supply_schedule = np.zeros(self.op_horizon)
        for i in range(self.op_horizon):
            if schedule[i] > 0:
                demand_schedule[i] = schedule[i]
            else:
                supply_schedule[i] = schedule[i]
        return [demand_schedule, supply_schedule]

    def plot_demand_profile(self, plot_now=False):
        self._get_heat_device_schedules()
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.FLPower, label='Fixed Load', linestyle='dashed')
        plt.plot(self._extract_demand_and_supply(self.EVSchedule)[0], label='EV charge', linestyle='dashed')
        plt.plot(self._extract_demand_and_supply(self.BatSchedule)[0], label='BAT charge', linestyle='dashed')
        plt.plot(self.DLSchedule, label='Deferrable Load', linestyle='dashed')
        plt.plot(self.EHSchedule, label='EH', linestyle='dashed')
        plt.plot(self.HPSchedule, label='HP', linestyle='dashed')
        self.final_demand_schedule = self.FLPower + self._extract_demand_and_supply(self.EVSchedule)[0] + \
                                     self._extract_demand_and_supply(self.BatSchedule)[0] + self.DLSchedule + \
                                     self.EHSchedule + self.HPSchedule
        plt.plot(self.final_demand_schedule, label='total')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Power Demand Profile')
        plt.grid()
        if plot_now:
            plt.show()
        return

    def plot_generation_profile(self, plot_now=False):
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(-self.PVPower, label='PV', linestyle='dashed')
        plt.plot(self._extract_demand_and_supply(self.EVSchedule)[1], label='EV discharge', linestyle='dashed')
        plt.plot(self._extract_demand_and_supply(self.BatSchedule)[1], label='BAT discharge', linestyle='dashed')
        plt.plot(self.CHPSchedule, label='CHP', linestyle='dashed')
        self.final_generation_schedule = -self.PVPower + self._extract_demand_and_supply(self.EVSchedule)[1] + \
                                     self._extract_demand_and_supply(self.BatSchedule)[1] + self.CHPSchedule
        plt.plot(self.final_generation_schedule, label='total')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Power Generation Profile')
        plt.grid()
        if plot_now:
            plt.show()
        return

    def complete_post_analyze(self):
        self.plot_flexible_device_schedules()
        self.plot_demand_profile()
        self.plot_generation_profile()
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.final_demand_schedule, label='Total demand', linestyle='dashed')
        plt.plot(self.final_generation_schedule, label='Total generation', linestyle='dashed')
        plt.plot(self.final_generation_schedule+self.final_demand_schedule, label= 'Total Imbalance')
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Electrical Imbalance Profile')
        plt.grid()
        #plt.show()
        return

    def print_schedules(self):
        return

    def _get_entity_schedules(self, entity_class):
        """
        Function that returns all schedules of one type of entity, e.g. Batteries
        """
        entity_list = []
        index = 0
        for en in self.city_district.get_all_entities():
            if entity_class == Battery and isinstance(en, Battery) and not isinstance(en, ElectricVehicle):
                schedule = self.convert_pyomo_to_numpy(en.model.p_el_vars)
                entity_list.append([])
                entity_list[index].append(en)
                entity_list[index].append(schedule)
                index += 1
            elif entity_class != Battery and isinstance(en, entity_class):
                schedule = self.convert_pyomo_to_numpy(en.model.p_el_vars)
                entity_list.append([])
                entity_list[index].append(en)
                entity_list[index].append(schedule)
                index += 1

        return entity_list

    def plot_entity_schedules(self, entity_class):
        """
        Function that plots all schedules of one type of entity, e.g. Batteries
        """
        entity_list = self._get_entity_schedules(entity_class)
        for i in range(len(entity_list)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(entity_list[i][1])
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title(entity_list[i][0])
            plt.grid()
        return

    def plot_city_district_schedule(self):
        """
        Function that plots the city districts' final schedule
        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(self.city_district.p_el_schedule)
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('City District Schedule')
        plt.grid()
        return

    def _compare_schedules(self, schedule_a, schedule_b):
        """
        Function that returns the average absolute deviation between two schedules in kW per time step
        """
        sum = 0
        for i in range(self.op_horizon):
            diff = schedule_a[i] - schedule_b[i]
            if diff < 0:
                diff = diff*(-1)
            print(diff)
            sum += diff
        average_deviation = sum / self.op_horizon
        return average_deviation

    def save_schedules(self):
        self._get_heat_device_schedules()
        self._get_flexible_device_schedules()
        self.district_schedule = np.add(self.city_district.p_el_schedule, np.zeros(self.op_horizon))
        return

    def compare_aggregated_entities(self, first_optimization,
                                    first_algorithm_name, second_algorithm_name, save_plots=False,):
        """
        Function to compare the aggregated schedules of each entity type, if different algorithms were used for the
        optimization of the same city district.
        It is mandatory that both optimizations are started in the same script and that an own DistrictAnalyzer object
        is created for each optimization. After the first executed optimization, the DistrictAnalyzer class method
        save_aggregated_schedules() has to be called to save the schedules for the later comparing.
        """
        if not np.array_equal(self.CHPSchedule, np.zeros(self.op_horizon)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(first_optimization.CHPSchedule, label=first_algorithm_name)
            plt.plot(self.CHPSchedule, label=second_algorithm_name)
            plt.legend()
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title('Aggregated CHP-Schedule Comparison')
            plt.grid()
            if save_plots:
                plt.savefig('aggregated_chp_schedules.png')

        if not np.array_equal(self.HPSchedule, np.zeros(self.op_horizon)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(self.HPSchedule, label=second_algorithm_name)
            plt.plot(first_optimization.HPSchedule, label=first_algorithm_name)
            plt.legend()
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title('Aggregated HP-Schedule Comparison')
            plt.grid()
            if save_plots:
                plt.savefig('aggregated_hp_schedules.png')

        if not np.array_equal(self.EHSchedule, np.zeros(self.op_horizon)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(self.EHSchedule, label=second_algorithm_name)
            plt.plot(first_optimization.EHSchedule, label=first_algorithm_name)
            plt.legend()
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title('Aggregated EH-Schedule Comparison')
            plt.grid()
            if save_plots:
                plt.savefig('aggregated_eh_schedules.png')

        if not np.array_equal(self.DLSchedule, np.zeros(self.op_horizon)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(self.DLSchedule, label=second_algorithm_name)
            plt.plot(first_optimization.DLSchedule, label=first_algorithm_name)
            plt.legend()
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title('Aggregated DL-Schedule Comparison')
            plt.grid()
            if save_plots:
                plt.savefig('aggregated_dl_schedules.png')

        if not np.array_equal(self.EVSchedule, np.zeros(self.op_horizon)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(self.EVSchedule, label=second_algorithm_name)
            plt.plot(first_optimization.EVSchedule, label=first_algorithm_name)
            plt.legend()
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title('Aggregated EV-Schedule Comparison')
            plt.grid()
            if save_plots:
                plt.savefig('aggregated_ev_schedules.png')
        return

    def compare_city_ditrict_schedules(self, first_optimization, first_algorithm_name, second_algorithm_name,
                                       save_plots=False):
        """
        Function that compares the final city district schedules if two different optimization algorithms were used.

        """
        plt.figure(DistrictAnalyzer.figure_number)
        DistrictAnalyzer.figure_number += 1
        plt.plot(first_optimization.district_schedule, label=first_algorithm_name)
        plt.plot(self.city_district.p_el_schedule, label=second_algorithm_name)
        plt.legend()
        plt.xlabel("Time step")
        plt.ylabel("Power in kW")
        plt.title('Compare Final Schedules')
        plt.grid()
        if save_plots:
            plt.savefig('compare_district_schedules.png')
        return

    def compare_entity_schedules(self, entity_class, first_optimization, first_algorithm_name, second_algorithm_name):
        first_entity_list = first_optimization._get_entity_schedules(entity_class)
        second_entity_list = self._get_entity_schedules(entity_class)
        for i in range(len(first_entity_list)):
            plt.figure(DistrictAnalyzer.figure_number)
            DistrictAnalyzer.figure_number += 1
            plt.plot(first_entity_list[i][1], label=first_algorithm_name)
            plt.plot(second_entity_list[i][1], label=second_algorithm_name)
            plt.xlabel("Time step")
            plt.ylabel("Power in kW")
            plt.title(first_entity_list[i][0])
            plt.grid()
        return

    def _get_constraints(self):
        equality_list = []
        inequality_list = []
        counter = 0
        for en in self.city_district.get_all_entities():
            for constraint in en.model.component_objects(pyomo.Constraint):
                counter += 1
                for index in constraint:
                    # check if the constraint is an equality constraint and write it in the form Ax-b=0
                    if pyomo.value(constraint[index].lower) == pyomo.value(constraint[index].upper):
                        expr = constraint[index].body - constraint[index].lower
                        equality_list.append([constraint, expr])

                    # if the constraint is not an equality constraint it has to be an inequality constraint
                    # the next three checks are about to write that constraint in the form Cx-d >= 0
                    elif pyomo.value(constraint[index].upper) is None:
                        expr = constraint[index].body - constraint[index].lower
                        inequality_list.append([constraint, expr])

                    elif pyomo.value(constraint[index].lower) is None:
                        expr = -constraint[index].body + constraint[index].upper
                        inequality_list.append([constraint, expr])

                    else:
                        expr = -constraint[index].body + constraint[index].upper
                        inequality_list.append([constraint, expr])
                        expr = constraint[index].body - constraint[index].lower
                        inequality_list.append([constraint, expr])

        return equality_list, inequality_list

    def count_violated_constraints(self):
        equality_list, inequality_list = self._get_constraints()
        equality_counter = 0
        inequality_counter = 0
        # Check if constraints are violated: allow a numerical tolerance of 1e-2 which equals 10 Watt in the application
        for i in range(len(equality_list)):
            if abs(pyomo.value(equality_list[i][1])) > 1e-2:
                print("Equality violation: ", equality_list[i][0])
                print(equality_list[i][1], "   ", pyomo.value(equality_list[i][1]))
                print("")
                equality_counter += 1
        for i in range(len(inequality_list)):
            if pyomo.value(inequality_list[i][1]) < -1e-2:
                inequality_counter += 1
                print("Inequality violation: ", inequality_list[i][0])
                print(inequality_list[i][1], "   ", pyomo.value(inequality_list[i][1]))
                print("")

        print("Number of inequality constraints:", len(inequality_list), " Number of violations: ", inequality_counter)
        print("Number of equality constraints:", len(equality_list), " Number of violations: ", equality_counter)
        return

    def check_start_vars_DL(self):
        for en in self.city_district.get_all_entities():
            if isinstance(en, DeferrableLoad):
                for variable in en.model.component_objects(pyomo.Var):
                    if variable[0] == 0 or variable[0] == 1:
                        variable.pprint()
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes.optimization_entity import OptimizationEntity
from pycity_scheduling import constants

class ThermalEntityCooling(OptimizationEntity):
    """
    Base class for all thermal cooling entities derived from OptimizationEntity.

    This class provides functionality common to all thermal cooling entities.


    Notes
    -----
    - Cooling TEs add the :math:`p_{th\\_cool}` variable to the model. When not modified
      by other classes, the following constraint is added:

    .. math::
        p_{th\\_cool} \\geq 0
    """

    def __init__(self, environment, *args, **kwargs):
        super().__init__(environment, *args, **kwargs)

        self.new_var("p_th_cool")

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel.

        Add variables for the thermal cooling demand of the entity to the optimization
        model.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer` : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model
        m.p_th_cool_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(-constants.INFINITY, constants.INFINITY),
                                     initialize=0)

        def p_th_cool_vars_domain_lb_rule(model, t):
            return model.p_th_cool_vars[t] >= -constants.INFINITY
        m.p_th_cool_vars_domain_lb_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_domain_lb_rule)

        def p_th_cool_vars_domain_ub_rule(model, t):
            return model.p_th_cool_vars[t] <= constants.INFINITY
        m.p_th_cool_vars_domain_ub_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_domain_ub_rule)
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *


# This is a very simple power scheduling example using the central optimization algorithm to demonstrate the impact
# of system level objective "valley-filling".


def main(do_plot=False):
    print("\n\n------ Example 16: Objective Valley-Filling ------\n\n")

    # Define timer, price, weather, and environment objects:
    t = Timer(op_horizon=96, step_size=900, initial_date=(2015, 4, 1))
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "valley-filling":
    valley_profile = [3.0 for i in range(24)] + [4.0 for i in range(24)] + [3.5 for i in range(48)]
    cd = CityDistrict(environment=e, objective='valley-filling', valley_profile=np.array(valley_profile))

    # Schedule some sample buildings. The buildings' objectives are defined as "none".
    n = 10
    for i in range(n):
        bd = Building(environment=e, objective='none')
        cd.addEntity(entity=bd, position=[0, i])
        bes = BuildingEnergySystem(environment=e)
        bd.addEntity(bes)
        ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
        bes.addDevice(ths)
        eh = ElectricHeater(environment=e, p_th_nom=10)
        bes.addDevice(eh)
        ap = Apartment(environment=e)
        bd.addEntity(ap)
        fi = FixedLoad(e, method=1, annual_demand=3000.0, profile_type='H0')
        ap.addEntity(fi)
        sh = SpaceHeating(environment=e, method=1, living_area=120, specific_demand=90, profile_type='HEF')
        ap.addEntity(sh)
        pv = Photovoltaic(environment=e, method=1, peak_power=8.2)
        bes.addDevice(pv)
        bat = Battery(environment=e, e_el_max=12.0, p_el_max_charge=4.6, p_el_max_discharge=4.6, eta=1.0)
        bes.addDevice(bat)

    # Perform the scheduling:
    opt = CentralOptimization(city_district=cd)
    results = opt.solve()
    cd.copy_schedule("valley-filling")

    # Print and show the city district's schedule:
    print("Schedule of the city district:")
    print(list(cd.p_el_schedule))

    gs = gridspec.GridSpec(3, 1)
    ax0 = plt.subplot(gs[0])
    ax0.plot(list(range(e.timer.timesteps_used_horizon)), cd.p_el_schedule)
    plt.ylim([-2.0, 5.0])
    plt.grid()
    plt.ylabel('Electrical power in kW')
    plt.title('City district scheduling result')

    ax1 = plt.subplot(gs[1], sharex=ax0)
    ax1.plot(list(range(e.timer.timesteps_used_horizon)), valley_profile)
    plt.grid()
    plt.ylabel('Reference power curve in kW')

    ax1 = plt.subplot(gs[2], sharex=ax0)
    ax1.plot(list(range(e.timer.timesteps_used_horizon)), np.array(cd.p_el_schedule) + np.array(valley_profile))
    plt.ylim([0.0, 10.0])
    plt.grid()
    plt.ylabel('Sum of both power curves in kW')

    plt.xlabel('Time in hours', fontsize=12)

    if do_plot:
        figManager = plt.get_current_fig_manager()
        if hasattr(figManager, "window"):
            figManagerWindow = figManager.window
            if hasattr(figManagerWindow, "state"):
                figManager.window.state("zoomed")
        plt.show()
    return


# Conclusions:
# Using "valley-filling" as the system level objective results in an "inverse" power profile for the considered city
# district compared to the "reference" power curve. The reference power curve usually represents a baseline with several
# power peaks and valleys that should get "compensated" taking advantage of the local flexibility potentials. In other
# words, this means that the sum of the city district's power profile and the reference power curve results in a "flat"
# power profile. This is usually the preferred system operation from the viewpoint of a network operator and/or district
# operator.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo

from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling
from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating
from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling import constants


class EntityContainer(ThermalEntityCooling, ThermalEntityHeating, ElectricalEntity):
    """
    Base class for entities containing other entities.

    `p_th` and `p_el` imbalances are propagated to this entities variables.
    During calls to its scheduling functions, the contained entities are also
    called with the same parameters.

    Notes
    -----
    - EntityContainers offer sets of constraints for operation. The following
      constraints are added.

    .. math::
        p_{th\\_cool} &=& \\sum_i p_{th\\_cool\\_i} \\\\
        p_{th\\_heat} &=& \\sum_i p_{th\\_heat\\_i} \\\\
        p_{el} &=& \\sum_i p_{el\\_i}

    - :math:`p_{th\\_cool\\_i}`, :math:`p_{th\\_heat\\_i}`, and :math:`p_{el\\_i}` are the variables from lower
      entities. The Bounds from TEC, TEH, and EE are removed.
    """

    def populate_model(self, model, mode="convex"):
        """
        Add entity block and lower entities blocks to pyomo ConcreteModel.

        Call both parent's `populate_model` methods and set variables lower
        bounds to `None`. Then call `populate_model` method of all contained
        entities and add constraints that the sum of their variables for each
        period equals the corresponding own variable.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model

        if mode in ["convex", "integer"]:
            p_th_cool_var_list = []
            p_th_heat_var_list = []
            p_el_var_list = []
            for entity in self.get_lower_entities():
                entity.populate_model(model, mode)
                if isinstance(entity, ThermalEntityCooling):
                    p_th_cool_var_list.append(entity.model.p_th_cool_vars)
                if isinstance(entity, ThermalEntityHeating):
                    p_th_heat_var_list.append(entity.model.p_th_heat_vars)
                if isinstance(entity, ElectricalEntity):
                    p_el_var_list.append(entity.model.p_el_vars)

            def p_th_cool_sum_rule(model, t):
                if len(p_th_cool_var_list) > 0:
                    return model.p_th_cool_vars[t] == sum(p_th_cool_var[t] for p_th_cool_var in p_th_cool_var_list)
                else:
                    return model.p_th_cool_vars[t] == 0.0
            m.p_th_cool_constr = pyomo.Constraint(m.t, rule=p_th_cool_sum_rule)

            def p_th_heat_sum_rule(model, t):
                if len(p_th_heat_var_list) > 0:
                    return model.p_th_heat_vars[t] == sum(p_th_heat_var[t] for p_th_heat_var in p_th_heat_var_list)
                else:
                    return model.p_th_heat_vars[t] == 0.0
            m.p_th_heat_constr = pyomo.Constraint(m.t, rule=p_th_heat_sum_rule)

            def p_el_sum_rule(model, t):
                if len(p_el_var_list) > 0:
                    return model.p_el_vars[t] == sum(p_el_var[t] for p_el_var in p_el_var_list)
                else:
                    return model.p_el_vars[t] == 0.0
            m.p_el_constr = pyomo.Constraint(m.t, rule=p_el_sum_rule)
        else:
            raise ValueError(
                "Mode %s is not implemented by class EntityContainer." % str(mode)
            )
        return

    def update_model(self, mode=""):
        super().update_model(mode)
        for entity in self.get_lower_entities():
            entity.update_model(mode)
        return

    def update_schedule(self):
        super().update_schedule()
        for entity in self.get_lower_entities():
            entity.update_schedule()
        return

    def reset(self, schedule=None):
        super().reset(schedule)
        for entity in self.get_lower_entities():
            entity.reset(schedule)
        return

    def get_lower_entities(self):
        raise NotImplementedError



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import matplotlib.pyplot as plt
import networkx as nx

from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.classes.electrical_node import ElectricalNode
from pycity_scheduling.classes.electrical_line import ElectricalLine


class ElectricalGrid(ElectricalEntity):
    """
    Class ElectricalGrid for scheduling purposes. Connects buildings within pycity_scheduling via a shared electrical
    grid.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
        Common to all other objects. Includes time and weather instances.
    related_entity : CityDistrict
        City district object that 'maintains' the ElectricalGrid.
    max_p_slack : float
        Maximum active power that can be exchanged through slack node of the grid in kW.
    max_q_slack : float
        Maximum reactive power that can be exchanged through slack node of the grid in kVar.
    slack_V : float, optional
        If not `None`, voltage at which the slack node should be operated in V. Default is None.
    ref_V : float, optional
        Nominal voltage in the electrical grid in V. Default is 400V.
    V_min : float, optional
        Minimal per unit voltage allowed in the electrical grid. Default is 0.95.
    V_max : float, optional
        Maximal per unit voltage allowed in the electrical grid. Default is 1.05.
    """
    def __init__(self, environment, related_entity, max_p_slack, max_q_slack, slack_V=None, ref_V=400, min_V=0.95,
                 max_V=1.05, *args, **kwargs):
        super().__init__(environment, *args, **kwargs)
        self._kind = "electricalgrid"
        self._long_id = "EG_" + self._id_string
        self.environment = environment
        self.related_entity = related_entity

        self.lines = []
        self.nb_lines = 0

        self.slack_node = ElectricalNode(environment, None, None)
        self.slack_V = slack_V
        self.nodes = dict()
        self.nodes[self.slack_node] = 0
        self.nb_nodes = 1

        self.connected_entities = dict()
        self.connected_entities[self.slack_node] = self.slack_node

        self.min_V = min_V
        self.max_V = max_V
        self.max_p_slack = max_p_slack
        self.max_q_slack = max_q_slack

        # Per unit calculations:
        self.ref_V = ref_V  # V
        self.ref_S = max_p_slack  # kVA
        self.ref_Z = self.ref_V ** 2 / (self.ref_S * 1000.0)

        # Cable properties per unit length:
        self.R_cable = 0.25  # Ohm/km
        self.X_cable = 0.0  # Ohm/km

    def add_node(self, node):
        """
        Add a node to the electrical grid.

        Check whether the node is already in the grid. If not, the node is assigned a number in the node dictionary and
        the node number is incremented by 1.
        If the node has a related entity, this entity is added to the connected_entity dictionary.
        The nodes inherit the voltage bounds of the electrical grid to which it is added.

        Parameters
        ----------
        node : ElectricalNode
            The node that should be added to the network.
        """
        if node not in self.nodes:
            node.min_V = self.min_V
            node.max_V = self.max_V
            if node.related_entity is not None:
                self.connected_entities[node.related_entity] = node
            self.nodes[node] = self.nb_nodes
            self.nb_nodes += 1

    def add_line(self, line):
        """
        Add a line to the electrical grid.
        
        Check whether the sending and receiving nodes of the line are already in the electrical grid. If not they are
        added. The line is added to the lines list of the grid and the line number is incremented by 1.

        Parameters
        ----------
        line : ElectricalLine
            The line that should be added to the network.
        """
        s = line.sending_node
        r = line.receiving_node
        self.add_node(s)
        self.add_node(r)
        self.lines.append(line)
        self.nb_lines += 1

    def remove_line(self, line):
        """
        Remove a line from the electrical grid.
        
        Remove the line from the line list and decrement the line number by 1.

        Parameters
        ----------
        line : ElectricalLine
            The line that should be removed from the network.
        """
        self.lines.remove(line)
        self.nb_lines -= 1

    def update_cable_properties(self, R_cable, X_cable):
        """
        Update the resistance and reactance cable parameters of the electrical grid.
        
        Update the R_cable (ohm/km) and X_cable (ohm/km) attributes of the grid with the provided values.

        Parameters
        ----------
        R_cable : float
            Cable resistance in ohm/km.
        X_cable : float
            Cable reactance in ohm/km.
        """
        self.R_cable = R_cable
        self.X_cable = X_cable

    def connect_entities(self, entity1, entity2, line_capacity, line_length, R_line_per_km=None, X_line_per_km=None):
        """
        Add a line of specified length and capacity linking two entities in an electrical grid.
        
        Convert line capacity in per unit according to the electrical grid nominal power.
        Calculate per unit resistance and reactance of the line using the grid cable properties.
        Add a line to the electrical grid to connect two entities, that can be electrical nodes or buildings.

        Parameters
        ----------
        entity1 : Union[ElectricalNode, Building]
            First entity to connect, sending node of the line.
        entity2 : Union[ElectricalNode, Building]
            Second entity to connect, receiving node of the line.
        line_capacity : float
            Line capacity in kVA.
        line_length : float
            Line length in km.
        """
        # Per unit conversion:
        line_capacity = line_capacity / self.ref_S

        # Impedance calculation based on cable properties and conversion in per unit:
        if R_line_per_km is None:
            R_line_per_km = self.R_cable
        if X_line_per_km is None:
            X_line_per_km = self.X_cable
        R_line = R_line_per_km * line_length / self.ref_Z
        X_line = X_line_per_km * line_length / self.ref_Z

        # Find electrical nodes related to the entities to be connected:
        if isinstance(entity1, ElectricalNode):
            node1 = entity1
        else:
            node1 = self.connected_entities[entity1]
        if isinstance(entity2, ElectricalNode):
            node2 = entity2
        else:
            node2 = self.connected_entities[entity2]

        # Add line to the electrical grid:
        line = ElectricalLine(self.environment, node1, node2, R_line, X_line, line_capacity)
        self.add_line(line)

    def build_incidence_matrix(self):
        """
        Build the incidence matrix of the electrical grid.

        Returns
        -------
        A : np.ndarray
            Incidence matrix of the grid.
        """
        A = np.zeros((self.nb_lines, self.nb_nodes))
        for i, line in enumerate(self.lines):
            s = self.nodes[line.sending_node]
            r = self.nodes[line.receiving_node]
            A[i, s] = 1
            A[i, r] = -1
        return A

    def populate_model(self, model, mode="convex"):
        """
        Add electrical grid block to pyomo ConcreteModel.

        Call parent's `populate_model` method. Then call `populate_model` method of all contained electrical lines and
        electrical nodes. Add constraints on slack node operation; bounded exchange power and enforcement of specified
        slack voltage if any. Add coupling constraints between nodes and their related entities.
        Add power conservation constraint at each node.
        
        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)

        # Populate nodes:
        for node in self.nodes:
            node.populate_model(model)
        # Populate lines:
        for line in self.lines:
            line.populate_model(model)

        m = self.model
        node_index = pyomo.RangeSet(0, self.nb_nodes-1)
        nodes_list = list(self.nodes.keys())

        # Slack node operation constraints
        # Apply specified slack voltage if any provided:
        if self.slack_V is not None:
            def slack_V_rule(model, t):
                return self.slack_node.model.V_el_vars[t] == self.slack_V ** 2
            model.slack_V_constr = pyomo.Constraint(m.t, rule=slack_V_rule)

        # Bounded injection power at slack node for healthy operation:
        self.slack_node.model.p_el_vars.setlb(- self.max_p_slack)
        self.slack_node.model.p_el_vars.setub(self.max_p_slack)
        self.slack_node.model.q_el_vars.setlb(- self.max_q_slack)
        self.slack_node.model.q_el_vars.setub(self.max_q_slack)

        # Coupling constraint between buildings and nodes power:
        def p_linking_rule(model, t, n):
            node = nodes_list[n]
            if node.related_entity is not None:
                return node.model.p_el_vars[t] + node.related_entity.model.p_el_vars[t] / self.ref_S == 0
            elif self.nodes[node] == 0:
                return m.p_el_vars[t] == 0
            else:
                return node.model.p_el_vars[t] == 0
        model.p_linking_constr = pyomo.Constraint(m.t, range(0, self.nb_nodes), rule=p_linking_rule)

        def q_linking_rule(model, t, n):
            node = nodes_list[n]
            if self.nodes[node] != 0:
                return node.model.q_el_vars[t] == node.ratio_pq_power * node.model.p_el_vars[t]
            else:
                return m.p_el_vars[t] == 0
        model.q_linking_constr = pyomo.Constraint(m.t, range(0, self.nb_nodes), rule=q_linking_rule)

        # Power flow constraint
        # Build incidence matrix:
        A = self.build_incidence_matrix()

        # Modified matrix for handling of losses in power conservation:
        A_loss = A.copy()
        A_loss[A_loss == -1] = 0

        # Enforce power flow conservation at each node, power losses are taken into account:
        def p_flow_rule(model, t, n):
            return sum(A[i, n] * line.model.p_el_vars[t] + A_loss[i, n] * line.model.p_loss_el_vars[t]
                       for i, line in enumerate(self.lines)) == nodes_list[n].model.p_el_vars[t]
        m.p_flow_constr = pyomo.Constraint(m.t, node_index, rule=p_flow_rule)

        def q_flow_rule(model, t, n):
            return sum(A[i, n] * line.model.q_el_vars[t] + A_loss[i, n] * line.model.q_loss_el_vars[t]
                       for i, line in enumerate(self.lines)) == nodes_list[n].model.q_el_vars[t]
        m.q_flow_constr = pyomo.Constraint(m.t, node_index, rule=q_flow_rule)

    def plot_graph_results(self, t):
        """
        Plot the optimization results for the electrical grid in an oriented graph for one time step.
        
        Represented electrical nodes of the electrical grid as graph nodes and lines as edges.
        Display the optimal values of power flow, power injection or generation and node voltages in per unit.

        Parameters
        ----------
        t : int
            Time step for which results are plotted.
        """
        # Create the incidence matrix of the grid:
        A = self.build_incidence_matrix()
        # Create a directed graph from the incidence matrix using the networkx package:
        G = nx.DiGraph()
        edge_labels = {}

        # Collect optimal values of power flows through lines:
        p_l_opti = []
        q_l_opti = []
        for line in self.lines:
            p_l_opti.append(line.model.p_el_vars[t].value * self.ref_S)
            q_l_opti.append(line.model.q_el_vars[t].value * self.ref_S)

        # Collect optimal values of power injection or generation and node voltages:
        p_n_opti = []
        q_n_opti = []
        V_opti = []
        for node in self.nodes:
            p_n_opti.append(node.model.p_el_vars[t].value * self.ref_S) 
            q_n_opti.append(node.model.q_el_vars[t].value * self.ref_S)
            V_opti.append(node.model.V_el_vars[t].value**0.5)

        # Collect optimal slack power injection:
        p_slack_opti = self.slack_node.model.p_el_vars[t].value * self.ref_S
        q_slack_opti = self.slack_node.model.q_el_vars[t].value * self.ref_S

        # Build graph
        # Add nodes and edges based on the incidence matrix and write edge label including optimal power flows:
        for i in range(A.shape[0]):
            for j in range(A.shape[1]):
                if A[i, j] == 1:
                    s = j
                elif A[i, j] == -1:
                    r = j
            edge_label = format(p_l_opti[i], '.3f') + " + j " + format(q_l_opti[i], '.3f')
            edge_labels[f'Node_{s}', f'Node_{r}'] = edge_label
            G.add_edge(f'Node_{s}', f'Node_{r}')

        # Draw the graph:
        plt.figure(figsize=(15, 10))
        pos = nx.spring_layout(G)
        nx.nx.draw_networkx(G, pos, with_labels=True, node_size=1700, node_color='skyblue', font_size=10,
                            font_color='black')
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_color='red')

        # Add node labels with power injection or generation and node voltages:
        for i in range(self.nb_nodes):
            if i == self.nodes[self.slack_node]:
                colour = "blue"
                node_label = ("S_slack: " + format(p_slack_opti, '.3f') + " + j " + format(q_slack_opti, '.3f') +
                              "\n V_slack: " + format(V_opti[i], '.3f'))
            else:
                colour = "black"
                node_label = ("S: " + format(p_n_opti[i], '.3f') + " + j " + format(q_n_opti[i], '.3f') +
                              "\n V: " + format(V_opti[i], '.3f'))
            x, y = pos[f'Node_{i}']
            plt.text(x + 0.05, y, node_label, fontsize=10, color=colour)

        plt.show()



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *

# This is a very simple power scheduling example using the distributed dual decomposition algorithm.


def main(do_plot=False):
    print("\n\n------ Example 04: Algorithm Dual-Decomposition ------\n\n")

    # Define timer, price, weather, and environment objects:
    t = Timer(op_horizon=2, step_size=3600)
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "peak-shaving":
    cd = CityDistrict(environment=e, objective='peak-shaving')

    # Schedule two sample buildings. The buildings' objectives are defined as "price".

    # Building no. one comes with fixed load, space heating, electric heater, pv unit, thermal energy storage, and
    # electrical energy storage:
    bd1 = Building(environment=e, objective='price')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(environment=e)
    bd1.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
    bes.addDevice(ths)
    eh = ElectricHeater(environment=e, p_th_nom=10)
    bes.addDevice(eh)
    ap = Apartment(environment=e)
    bd1.addEntity(ap)
    load = np.array([10.0, 10.0])
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    pv = Photovoltaic(environment=e, method=1, peak_power=4.6)
    bes.addDevice(pv)
    bat = Battery(environment=e, e_el_max=4.8, p_el_max_charge=3.6, p_el_max_discharge=3.6, eta=1.0)
    bes.addDevice(bat)

    # Building no. two comes with deferrable load, curtailable load, space heating, chp unit, thermal energy storage,
    # and an electric vehicle:
    bd2 = Building(environment=e, objective='price')
    cd.addEntity(entity=bd2, position=[0, 1])
    bes = BuildingEnergySystem(environment=e)
    bd2.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=35, soc_init=0.5)
    bes.addDevice(ths)
    chp = CombinedHeatPower(environment=e, p_th_nom=20.0)
    bes.addDevice(chp)
    ap = Apartment(environment=e)
    bd2.addEntity(ap)
    load = np.array([20.0, 20.0])
    dl = DeferrableLoad(environment=e, p_el_nom=2.0, e_consumption=2.0, load_time=[1, 1])
    ap.addEntity(dl)
    cl = CurtailableLoad(environment=e, p_el_nom=1.6, max_curtailment=0.8)
    ap.addEntity(cl)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    ev = ElectricVehicle(environment=e, e_el_max=37.0, p_el_max_charge=22.0, soc_init=0.1, charging_time=[1, 1],
                         simulate_driving=False, minimum_soc_end=0.9)
    ap.addEntity(ev)

    # Perform the scheduling:
    opt = DualDecomposition(city_district=cd, rho=0.1, eps_primal=1.0, max_iterations=1000)
    results = opt.solve()
    cd.copy_schedule("dual_decomposition")

    # Print some ADMM results:
    print("Dual Decomposition - Number of iterations:")
    print(results["iterations"][-1])
    print("Dual Decomposition - Norm vector 'r' over iterations:")
    print(results["r_norms"])
    print("Dual Decomposition - Final shadow price vector:")
    print(results["lambdas"][-1])
    print("")

    # Print the building's schedules:
    print("Schedule building no. one:")
    print(list(bd1.p_el_schedule))
    print("Schedule building no. two:")
    print(list(bd2.p_el_schedule))
    print("Schedule of the city district:")
    print(list(cd.p_el_schedule))
    print("Objective_value: ", results["obj_value"][-1])
    return


# Conclusions:
# If the distributed dual decomposition optimization algorithm is applied, the two buildings are scheduled in a way
# so that both the local and system level objectives are satisfied. Local flexibility is used to achieve the system
# level objective. The scheduling results are close to the ones of the central algorithm, which demonstrates the
# correctness of the distributed algorithm. Nevertheless, Dual Decomposition can be slow in convergence!


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)
